package pnews.servlet;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jsoup.Jsoup;

import com.rometools.rome.feed.synd.SyndEnclosure;
import com.rometools.rome.feed.synd.SyndEntry;
import com.rometools.rome.feed.synd.SyndFeed;
import com.rometools.rome.io.FeedException;
import com.rometools.rome.io.SyndFeedInput;
import com.rometools.rome.io.XmlReader;

import pnews.Article;
import pnews.Category;

public class ArticleProvider {
        public final static ArticleProvider singleton = new ArticleProvider();
        private static final Logger LOG = Logger.getLogger(ArticleProvider.class.getName());
        private final Map<Category, List<Article>> articlesByCategory = new HashMap<>();
        private final ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(2);
        
        private ArticleProvider() {      
                for (Category cat:Category.values())
                        scheduler.scheduleAtFixedRate(new Refresher(cat), 2, 600, TimeUnit.SECONDS);
        }
        
        private static SyndFeed getSyndFeed(String u) throws IllegalArgumentException, FeedException, MalformedURLException, IOException {
                XmlReader r;
                
                r = new XmlReader(new URL(u));
                
                return new SyndFeedInput().build(r);                
        }
        
        private static Map<Category, String[]> getFeeds() {
                Map<Category, String[]> result;
                
                result = new HashMap<>();
                
                result.put(Category.TOP,
                           new String[] {
                                           "http://www.francetvinfo.fr/titres.rss",
                                           "http://www.rfi.fr/general/rss",
                                           "http://www.cnews.fr/rss/une",
                                           "http://www.ladepeche.fr/rss/a-la-une.rss",
                                           "https://www.franceinter.fr/rss/a-la-une.xml",
                                           "https://www.francebleu.fr/rss/a-la-une.xml",
                                           "http://www.bfmtv.com/rss/info/flux-rss/flux-toutes-les-actualites/"
                           });
                
                result.put(Category.SPORT,
                                new String[] { "http://www.france24.com/fr/sports/rss" });
                
                result.put(Category.FRANCE,
                                new String[] { "http://www.france24.com/fr/france/rss",
                                               "http://www.francetvinfo.fr/france.rss",
                                               "http://www.rfi.fr/france/rss"});
                
                result.put(Category.EUROPE,
                                new String[] { "http://www.france24.com/fr/europe/rss" });

                result.put(Category.MONDE, 
                           new String[] { "http://www.france24.com/fr/actualites/rss" });                                           

                
                result.put(Category.ECO,
                                new String[] { "http://www.france24.com/fr/economie/rss",
                                               "http://www.rfi.fr/economie/rss" });
                
                result.put(Category.ESSONNE,
                                new String[] { "http://www.tourisme-essonne.com/rss/actus/",
                                               "http://www.ville-palaiseau.fr/rss/actualites.htm" });
                
                result.put(Category.PEOPLE,
                                new String[] { "http://www.premiere.fr/rss/actu-live",
                                               "http://www.purepeople.com/rss/news_t0.xml"                                               
                });
                
                result.put(Category.TECHNOLOGIE,
                                new String[] { "http://feeds.feedburner.com/lesnumeriques/news",
                                               "http://www.zdnet.fr/feeds/rss/actualites/",
                                               "http://www.frandroid.com/feed",
                                               "http://www.silicon.fr/feed",
                                               "http://www.fredzone.org/feed",
                                               "http://www.futura-sciences.com/rss/actualites.xml",
                                               "https://www-03.ibm.com/press/fr/fr/rssfeed.wss?keyword=null&maxFeed=&feedType=RSS&topic=all"});
                
                return result;
        }
        
        private List<Article> getArticlesForUpdate(Category cat) {
                List<Article> result;
                
                synchronized (articlesByCategory) {
                        result = articlesByCategory.get(cat);
                        if (result == null) {
                                result = new ArrayList<>();
                                articlesByCategory.put(cat, result);
                        }
                        return result;
                }                
        }
        
        private boolean exists(String articleLink, List<Article> articles) {
                synchronized (articles) {
                        for (Article a: articles)
                                if (a.link.equals(articleLink))
                                        return true;
                }
                return false;
        }
        
        private static Article toArticle(String link, SyndEntry entry, SyndFeed feed) {
                String desc, title, thumbnail, feedTitle, str;
                Date date;
                
                feedTitle = feed.getTitle();
                if (feedTitle != null) {
                        feedTitle = feedTitle.trim();
                }
                
                thumbnail = null;
                for (SyndEnclosure e: entry.getEnclosures()) {
                        if (e.getType().startsWith("image/"))
                                thumbnail = e.getUrl();    
                        break;
                }
                
                if (thumbnail == null && feed.getImage() != null)
                        thumbnail = feed.getImage().getUrl();
                             
                
                title = entry.getTitle().trim();
                
                if (entry.getDescription() != null) {
                        str = entry.getDescription().getValue();
                        desc = Jsoup.parse(str).text();
                } else {       
                        desc = null;
                        LOG.severe("No description for " + feedTitle + " - " + title);
                }
                
                date = entry.getPublishedDate();
                if (date == null)
                        date = entry.getUpdatedDate();
                if (date == null)
                        LOG.severe("The article " + feedTitle + " - " + title + " does not have a date");
                                     
                return new Article(link, title, desc, thumbnail, date, feedTitle);
        }
        
        private void addArticles(Category cat, SyndFeed feed) {
                String feedTitle;
                List<Article> articles;
                Article a;
                
                feedTitle = feed.getTitle().trim();
                
                LOG.info("addArticles " + cat.getId() + " " + feedTitle + " number of articles: " + feed.getEntries().size());
                
                for (SyndEntry entry: feed.getEntries()) {
                        String link = entry.getLink().trim();
                        articles = getArticlesForUpdate(cat);
                        if (exists(link, articles)) {
                                LOG.fine("addArticles " + link + " is already present");
                                continue ;
                        }
                        
                        a = ArticleStore.singleton.getArticle(link, ()->toArticle(link, entry, feed));
                        
                        synchronized (articles) {
                                articles.add(a);

                                Collections.sort(articles, new Comparator<Article>() {
                                        @Override
                                        public int compare(Article o1, Article o2) {
                                                if (o1.publicationDate == o2.publicationDate)
                                                        return 0;
                                                if (o1.publicationDate == null)
                                                        return 1;
                                                if (o2.publicationDate == null)
                                                        return -1;
                                                return o2.publicationDate.compareTo(o1.publicationDate);
                                        }
                                });
                        }
                }          
                
                LOG.info("addArticles done " + cat.getId());
        }
             
        private void retrieveArticles(Category cat) throws IllegalArgumentException, MalformedURLException, FeedException, IOException {
                String[] feeds;
                
                feeds = getFeeds().get(cat);
                
                if (feeds != null)
                        for (String str: feeds)
                                try {
                                        addArticles(cat, getSyndFeed(str));
                                } catch (Throwable e) {
                                        LOG.log(Level.SEVERE,
                                                "retrieveArticles failure " + cat.getId() + " " + str,
                                                e);
                                }
                else
                        LOG.severe("No feed for category " + cat);
        }
        
        /**
         * Returns a copy.
         */
        public List<Article> getArticles(Category cat)
                        throws IllegalArgumentException, MalformedURLException, FeedException, IOException {
                List<Article> articles;
                
                synchronized (articlesByCategory) {
                        articles = getArticlesForUpdate(cat);
                }
                
                synchronized (articles) {
                        return new ArrayList<>(articles);
                }
        }
        
        private class Refresher implements Runnable {
                private final Category category;
                
                public Refresher(Category category) {
                        this.category = category;
                }
                
                @Override
                public void run() {                       
                        LOG.info("refresher "+ category.getId());
                        
                        try {
                                retrieveArticles(category);
                        } catch (IllegalArgumentException | FeedException | IOException e) {
                                LOG.log(Level.SEVERE, "refresher failure", e);
                        }                        
                        
                        LOG.info("refresher "+ category.getId() + " done");
                }                
        }
}
